#include "config.h"
#include "boushi.h"
#include "sha1.h"

#include <stdbool.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <db.h>
#include <lzo/lzo1x.h>
#include <arpa/inet.h>

static const char db_config_defaults[] =
	"set_flags DB_LOG_AUTOREMOVE\n"
	"#set_cachesize 1 0 0\n"
	"#set_lg_bsize 33554432\n"
	"#set_lg_max 134217728\n"
	"#set_lg_dir /some/reliable/disk1\n"
	"#set_data_dir /some/reliable/disk2\n";

struct boushi {
	DB_ENV *env;
	DB *content;
	DB *files;
	DB *directories;
	char *dbdir;
	char *error;
	char *lzowork;
	int err;
};

static const boushi_t boushi_0 = {0};

size_t boushi_size(void) {
	return sizeof boushi_0;
}

static const DBT DBT0 = {0};

static const char *boushi_enomem = "Out of memory";

#define ok(err, ...) boushi_ok(b, err, __VA_ARGS__, NULL)
#define bok(...) boushi_ok(b, -1, __VA_ARGS__, NULL)

static bool boushi_ok(boushi_t *b, int err, ...) {
	va_list ap;
	char *msg, *errstr = NULL;
	char *s;
	size_t l, len;
	bool colon;

	if(!b)
		return !err;

	b->err = err;
	if(b->error) {
		if(b->error != boushi_enomem)
			free(b->error);
		b->error = NULL;
	}

	if(!err)
		return true;

	va_start(ap, err);
	colon = false;
	len = 0;
	while(s = va_arg(ap, char *), s) {
		if(colon)
			len += 2;
		else
			colon = true;
		len += strlen(s);
	}
	va_end(ap);

	if(err != -1) {
		if(colon)
			len += 2;
		errstr = db_strerror(err);
		if(!errstr)
			return false;
		len += strlen(errstr);
	}
		
	msg = malloc(len + 1);

	va_start(ap, err);
	colon = false;
	len = 0;
	while(s = va_arg(ap, char *), s) {
		if(colon) {
			memcpy(msg + len, ": ", 2);
			len += 2;
		} else {
			colon = true;
		}
		l = strlen(s);
		memcpy(msg + len, s, l);
		len += l;
	}
	va_end(ap);

	if(errstr) {
		if(colon) {
			memcpy(msg + len, ": ", 2);
			len += 2;
		}
		strcpy(msg + len, errstr);
	} else {
		msg[len] = '\0';
	}

	b->error = msg;

	return false;
}

int boushi_errno(struct boushi *b) {
	return b ? b->err : EFAULT;
}

const char *boushi_error(struct boushi *b) {
	return b ? b->error ? b->error : "" : "boushi object is NULL";
}

static size_t file2path(char *dst, const char *src) {
	char *cur, *end;
	const char *sep;
	size_t len;

#if 0
	if(*src == '/')
		*dst++ = *src++;
#endif

	cur = dst;
	do {
		/* strchrnul() */
		sep = strchr(src, '/');
		if(!sep)
			sep = src + strlen(src);
		len = sep - src;
		if(!len) {
			// do nothing
		} else if(len == 1 && *src == '.') {
			// do nothing
		} else if(len == 2 && src[0] == '.' && src[1] == '.') {
			end = memrchr(dst, '/', cur - dst);
			if(end)
				cur = end;
		} else {
			if(cur > dst)
				*cur++ = '/';
			memcpy(cur, src, len);
			cur += len;
		}
		src = sep + 1;
	} while(*sep);

	*cur = '\0';

	return cur - dst;
}

static int pathcmp(DB *db, const DBT *a, const DBT *b) {
	const char *ap, *bp, *ae, *be;
	size_t al, bl;
	int c;

	al = a->size;
	bl = b->size;

	if(!al)
		return bl ? -1 : 0;
	if(!bl)
		return 1;

	ap = a->data;
	bp = b->data;
	ae = ap + al;
	be = bp + bl;

	for(;;) {
		ap = memchr(ap, '/', ae - ap);
		bp = memchr(bp, '/', be - bp);
		if(ap) {
			if(!bp)
				return 1;
		} else {
			if(bp)
				return -1;
			break;
		}
		ap++;
		bp++;
	}

	c = memcmp(a->data, b->data, al < bl ? al : bl);
	if(c)
		return c;
	return al < bl ? -1 : al > bl ? 1 : 0;
}

#define SIZEBYTES 4
#define MAXSIZE ((size_t)1 << (SIZEBYTES << 3))

static uint32_t readsize(const uint8_t *buf) {
	return (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
}

static void writesize(uint8_t *buf, size_t size) {
	buf[0] = size >> 24;
	buf[1] = (size >> 16) & 255;
	buf[2] = (size >> 8) & 255;
	buf[3] = size & 255;
}

static bool lzo_pack(boushi_t *b, const uint8_t *in, size_t inlen, uint8_t **outp, size_t *outlenp) {
	lzo_uint lzolen = 0;
	uint8_t *out = NULL;
	static char *work = NULL;
	bool peachy = true;

	if(!inlen) {
		*outp = NULL;
		*outlenp = 0;
		return peachy;
	}

	if(inlen <= MAXSIZE) {
		work = b->lzowork;
		if(!work) {
			work = malloc(LZO1X_999_MEM_COMPRESS);
			if(work) {
				b->lzowork = work;
				memset(work, 0, LZO1X_999_MEM_COMPRESS);
			} else {
				peachy = ok(errno, "malloc");
			}
		}
		if(peachy) {
			out = malloc(inlen + inlen / 16 + 64 + 3 + SIZEBYTES);
			if(!out)
				peachy = ok(errno, "malloc");
		}
		if(peachy) {
			if(lzo1x_999_compress(in, inlen, out, &lzolen, work) != LZO_E_OK)
				peachy = bok("lzo1x_999_compress");
		}
		if(peachy && lzolen < inlen) {
			writesize(out + lzolen, inlen);
			*outp = out;
			*outlenp = lzolen + SIZEBYTES;
			return peachy;
		}
	}

	if(peachy && !out) {
		out = malloc(inlen + SIZEBYTES);
		if(!out)
			peachy = ok(errno, "malloc");
	}

	if(peachy) {
		writesize(out + inlen, 0);
		memcpy(out, in, inlen);

		*outp = out;
		*outlenp = inlen + SIZEBYTES;
		return peachy;
	}

	if(out)
		free(out);

	return peachy;
}

static bool lzo_unpack(boushi_t *b, uint8_t *buf, size_t size, uint8_t **out, size_t *outsize) {
	bool peachy = true;
	uint8_t *lzo = NULL;
	uint32_t lzosize = 0;
	lzo_uint reslen = 0;

	if(!buf || size < SIZEBYTES || !out || !outsize)
		return ok(EINVAL, "internal error");

	size -= SIZEBYTES;

	lzosize = readsize(buf + size);
	if(!lzosize) {
		buf[size] = '\0';
		*out = buf;
		*outsize = size;
		return true;
	}
	lzo = malloc(lzosize + 1);
	if(!lzo)
		return ok(errno, "malloc");

	if(lzo1x_decompress(buf, size, lzo, &reslen, NULL) != LZO_E_OK)
		peachy = false;
	if(reslen != lzosize)
		peachy = false;

	if(peachy) {
		lzo[lzosize] = '\0';
		*out = lzo;
		*outsize = lzosize;
	} else if(lzo) {
		free(lzo);
	}

	return peachy;
}

void boushi_exit(boushi_t *b) {
	if(!b)
		return;

	if(b->directories)
		b->directories->close(b->directories, 0);

	if(b->files)
		b->files->close(b->files, 0);

	if(b->content)
		b->content->close(b->content, 0);

	if(b->env)
		b->env->close(b->env, 0);

	if(b->error && b->error != boushi_enomem)
		free(b->error);

	if(b->dbdir)
		free(b->dbdir);

	if(b->lzowork)
		free(b->lzowork);

	*b = boushi_0;
}

static bool boushi_write_db_config(boushi_t *b) {
	int fd = -1, dirfd = -1;
	ssize_t size = sizeof db_config_defaults - 1;
	ssize_t r;
	bool peachy = true;

	if(peachy) {
		dirfd = open(b->dbdir, O_RDONLY|O_NOCTTY|O_DIRECTORY);
		if(dirfd == -1)
			peachy = ok(errno, "open", b->dbdir);
	}

	if(peachy) {
		fd = openat(dirfd, "DB_CONFIG", O_CREAT|O_EXCL|O_WRONLY|O_NOCTTY, 0666);
		if(fd == -1 && errno != EEXIST)
			peachy = ok(errno, "openat", b->dbdir, "DB_CONFIG");
	}

	if(dirfd != -1) {
		if(close(dirfd) == -1)
			peachy = ok(errno, "close", b->dbdir);
	}

	if(fd != -1) {
		if(peachy) {
			r = write(fd, db_config_defaults, size);
			if(r == -1)
				peachy = ok(errno, "write", "DB_CONFIG");
			if(r != size)
				peachy = ok(ECOMM, "write", "DB_CONFIG");
		}

		if(close(fd) == -1)
			peachy = ok(errno, "close", "DB_CONFIG");
	}

	return peachy;
}

static uint32_t boushi_hash(DB *db, const void *bytes, u_int32_t length) {
	switch(length) {
		case 0: return 0;
		case 1: return ((uint8_t *)bytes)[0] << 24;
		case 2: return ((uint8_t *)bytes)[0] << 24 | ((uint8_t *)bytes)[1] << 16;
		case 3: return ((uint8_t *)bytes)[0] << 24 | ((uint8_t *)bytes)[1] << 16 | ((uint8_t *)bytes)[2] << 8;
		default: return ((uint8_t *)bytes)[0] << 24 | ((uint8_t *)bytes)[1] << 16 | ((uint8_t *)bytes)[2] << 8 | ((uint8_t *)bytes)[3];
	}
}

static bool boushi_checkpoint(boushi_t *b) {
	uint32_t x;
	return ok(b->env->get_lg_max(b->env, &x), "get_lg_max")
		&& ok(b->env->txn_checkpoint(b->env, x/4096, 0, 0), "txn_checkpoint");
}

bool boushi_init(boushi_t *b, const char *dbdir) {
	DB_TXN *txn = NULL;
	bool peachy = true;

	if(!b)
		return false;

	*b = boushi_0;

	if(!dbdir)
		return bok("boushi_init", "dbdir is NULL");

	if(peachy) {
		b->dbdir = strdup(dbdir);
		peachy = b->dbdir || ok(errno, "malloc");
	}

	if(mkdir(dbdir, 0777) == -1)
		peachy = errno == EEXIST || ok(errno, "mkdir", dbdir);
	else
		peachy = boushi_write_db_config(b);

	if(peachy)
		peachy = ok(db_env_create(&b->env, 0), "db_env_create");

	if(peachy)
		peachy = ok(b->env->open(b->env, dbdir, DB_CREATE|DB_RECOVER|DB_REGISTER|DB_THREAD|DB_INIT_LOCK|DB_INIT_LOG|DB_INIT_MPOOL|DB_INIT_TXN|DB_USE_ENVIRON, 0666), "env_open", dbdir);

	if(peachy)
		peachy = ok(db_create(&b->content, b->env, 0), "db_create");

	if(peachy)
		peachy = ok(b->content->set_h_hash(b->content, boushi_hash), "set_h_hash");

	if(peachy)
		peachy = ok(db_create(&b->files, b->env, 0), "db_create");

	if(peachy)
		peachy = ok(db_create(&b->directories, b->env, 0), "db_create");

	if(peachy)
		peachy = ok(b->directories->set_bt_compare(b->directories, pathcmp), "set_bt_compare");

	if(peachy)
		peachy = ok(b->env->txn_begin(b->env, NULL, &txn, 0), "txn_begin");

	if(peachy)
		peachy = ok(b->content->open(b->content, txn, "content.hash", NULL, DB_HASH, DB_CREATE, 0666), "db->open", "content", "content");

	if(peachy)
		peachy = ok(b->files->open(b->files, txn, "directory.btree", "files", DB_BTREE, DB_CREATE, 0666), "db->open", "directory", "files");

	if(peachy)
		peachy = ok(b->directories->open(b->directories, txn, "directory.btree", "directories", DB_BTREE, DB_CREATE, 0666), "db->open", "directory", "directories");

	if(peachy)
		peachy = ok(txn->commit(txn, 0), "txn->commit");
	else if(txn)
		txn->abort(txn);

	if(peachy)
		peachy = boushi_checkpoint(b);

	return peachy;
}

void sha1(const char *s, size_t len, uint8_t *r) {
	struct sha1_ctxt ctx;

	if(!r)
		return;

	if(s) {
		sha1_init(&ctx);
		sha1_loop(&ctx, (const uint8_t *)s, len);
		sha1_result(&ctx, r);
	} else {
		memset(r, '\0', SHA1_RESULTLEN);
	}
}

bool boushi_put(boushi_t *b, const char *name, const void *buf, size_t len) {
	char *path = NULL;
	size_t pathlen = 0;
	uint8_t *out = NULL;
	size_t outlen = 0;
	DB_TXN *txn = NULL;
	DBT key = DBT0, val = DBT0;
	bool peachy = true;
	uint8_t sha1sum[SHA1_RESULTLEN];
	uint32_t refcount;
	bool found = false, isfile = true;
	int err;
	char *s;

	if(!name)
		return bok("boushi_put", "name parameter is NULL");

	if(!buf && len)
		return bok("boushi_put", "buf parameter is NULL");

	path = alloca(strlen(name) + 1);
	pathlen = file2path(path, name);

	sha1(path, pathlen, sha1sum);

	peachy = lzo_pack(b, buf, len, &out, &outlen);

	if(peachy && outlen > UINT32_MAX)
		peachy = bok("file too large", name);

	if(peachy)
		peachy = ok(b->env->txn_begin(b->env, NULL, &txn, 0), "txn_begin");

	if(peachy) {
		key.data = sha1sum;
		key.ulen =
		key.size = sizeof sha1sum;
		val.data = out;
		val.ulen =
		val.size = outlen;
		peachy = ok(b->content->put(b->content, txn, &key, &val, 0), "put", name);
	}

	if(out)
		free(out);

	if(peachy) {
		key.data = path;
		key.ulen =
		key.size = pathlen + 1;
		val.data = NULL;
		val.ulen =
		val.size = 0;
		peachy = ok(b->files->put(b->files, txn, &key, &val, 0), "put", name);
	}

	while(peachy) {
		key.data = path;
		key.ulen =
		key.size = pathlen + 1;
		val.data = &refcount;
		val.ulen = sizeof refcount;
		val.flags = DB_DBT_USERMEM;
		err = b->directories->get(b->directories, txn, &key, &val, DB_RMW);
		if(err) {
			if(err == DB_NOTFOUND) {
				refcount = 0;
			} else {
				peachy = ok(err, "get", path);
				break;
			}
		} else {
			if(val.size != sizeof refcount) {
				bok("database corrupted (needs dump+load)");
				break;
			}
			found = true;
			refcount = ntohl(refcount);
		}

		if(isfile)
			refcount |= UINT32_C(0x80000000);
		else
			refcount++;

		refcount = htonl(refcount);

		val.data = &refcount;
		val.ulen =
		val.size = sizeof refcount;
		peachy = ok(b->directories->put(b->directories, txn, &key, &val, 0), "put", path);
		if(!peachy || found)
			break;

		s = memrchr(path, '/', pathlen);
		if(!s)
			break;
		*s = '\0';
		pathlen = s - path;

		isfile = false;
	}

	if(peachy)
		peachy = ok(txn->commit(txn, 0), "txn->commit");
	else if(txn)
		txn->abort(txn);

	if(peachy)
		peachy = boushi_checkpoint(b);

	return peachy;
}

bool boushi_del(boushi_t *b, const char *name, bool *isfound) {
	char *path = NULL;
	size_t pathlen = 0;
	DB_TXN *txn = NULL;
	DBT key = DBT0, val = DBT0;
	bool peachy = true;
	uint8_t sha1sum[SHA1_RESULTLEN];
	uint32_t refcount;
	bool done = false, found = true, isfile = true;
	int err;
	char *s;

	if(!name)
		return bok("boushi_del", "name parameter is NULL");

	path = alloca(strlen(name) + 1);
	pathlen = file2path(path, name);

	sha1(path, pathlen, sha1sum);

	peachy = ok(b->env->txn_begin(b->env, NULL, &txn, 0), "txn_begin");

	if(peachy) {
		key.data = sha1sum;
		key.ulen =
		key.size = sizeof sha1sum;
		err = b->content->del(b->content, txn, &key, 0);
		if(err) {
			if(err == DB_NOTFOUND)
				found = false;
			else
				peachy = ok(err, "del", name);
		}
	}

	if(found) {
		if(peachy) {
			key.data = path;
			key.ulen =
			key.size = pathlen + 1;
			peachy = ok(b->files->del(b->files, txn, &key, 0), "del", name);
		}

		for(;;) {
			key.data = path;
			key.ulen =
			key.size = pathlen + 1;
			val.data = &refcount;
			val.ulen = sizeof refcount;
			val.flags = DB_DBT_USERMEM;
			peachy = ok(b->directories->get(b->directories, txn, &key, &val, DB_RMW), "get", path);
			if(!peachy)
				break;

			if(val.size != sizeof refcount) {
				bok("database corrupted (needs dump+load)");
				break;
			}

			refcount = ntohl(refcount);
			if(isfile) {
				if(!(refcount & UINT32_C(0x80000000))) {
					bok("database corrupted (needs dump+load)");
					break;
				}
				refcount &= ~UINT32_C(0x80000000);
			} else {
				if(!refcount) {
					bok("database corrupted (needs dump+load)");
					break;
				}
				refcount--;
			}
			refcount = htonl(refcount);

			if(refcount) {
				val.data = &refcount;
				val.ulen =
				val.size = sizeof refcount;
				peachy = ok(b->directories->put(b->directories, txn, &key, &val, 0), "put", path);
				done = true;
			} else {
				peachy = ok(b->directories->del(b->directories, txn, &key, 0), "del", path);
			}

			if(!peachy || done)
				break;

			s = memrchr(path, '/', pathlen);
			if(!s)
				break;
			*s = '\0';
			pathlen = s - path;

			isfile = false;
		}
	}

	if(peachy)
		peachy = ok(txn->commit(txn, 0), "txn->commit");
	else if(txn)
		txn->abort(txn);

	if(peachy && isfound)
		*isfound = found;

	if(peachy)
		peachy = boushi_checkpoint(b);

	return peachy;
}

bool boushi_has(boushi_t *b, const char *name, bool *isfound) {
	char *path = NULL;
	size_t pathlen = 0;
	DBT key = DBT0;
	uint8_t sha1sum[SHA1_RESULTLEN];
	int err;

	if(!name)
		return bok("boushi_has", "name parameter is NULL");

	path = alloca(strlen(name) + 1);
	pathlen = file2path(path, name);

	sha1(path, pathlen, sha1sum);

	key.data = sha1sum;
	key.ulen =
	key.size = sizeof sha1sum;
	err = b->content->exists(b->content, NULL, &key, 0);
	switch(err) {
		case 0:
			if(isfound)
				*isfound = true;
			return true;
		case DB_NOTFOUND:
			if(isfound)
				*isfound = false;
			return true;
		default:
			return ok(err, "has", name);
	}
}

bool boushi_get(boushi_t *b, const char *name, void **buf, size_t *len) {
	char *path = NULL;
	size_t pathlen = 0;
	uint8_t *out = NULL;
	size_t outlen = 0;
	DBT key = DBT0, val = DBT0;
	bool peachy = true;
	uint8_t sha1sum[SHA1_RESULTLEN];
	int err;

	if(!name)
		return bok("boushi_get", "name parameter is NULL");

	if(!buf && len)
		return bok("boushi_get", "buf parameter is NULL");

	path = alloca(strlen(name) + 1);
	pathlen = file2path(path, name);

	sha1(path, pathlen, sha1sum);

	key.data = sha1sum;
	key.ulen =
	key.size = sizeof sha1sum;
	val.flags = DB_DBT_MALLOC;
	err = b->content->get(b->content, NULL, &key, &val, 0);
	if(err == DB_NOTFOUND) {
		*buf = NULL;
		*len = 0;
		return true;
	}

	peachy = ok(err, "get", name);

	if(peachy)
		peachy = lzo_unpack(b, val.data, val.size, &out, &outlen);

	if(peachy && val.data == out)
		val.data = NULL;

	if(val.data)
		free(val.data);

	if(peachy) {
		*buf = out;
		*len = outlen;
	}

	return peachy;
}

struct boushi_cursor {
	boushi_t *b;
	char *next;
	char *prefix;
	size_t len;
	int err;
	bool recursive;
};

static const boushi_cursor_t boushi_cursor_0 = {0};

size_t boushi_cursor_size(void) {
	return sizeof boushi_cursor_0;
}

bool boushi_cursor_exit(boushi_cursor_t *c) {
	if(!c)
		return false;
	if(c->prefix)
		free(c->prefix);
	if(c->next)
		free(c->next);
	*c = boushi_cursor_0;
	return true;
}

static bool boushi_cursor_succ(boushi_t *b, DB *db, const char *s, char **o, size_t *olen) {
	DBC *cursor;
	DBT key = DBT0, val = DBT0;
	size_t len;
	int err;
	bool peachy = true;

	len = strlen(s);

	if(!ok(db->cursor(db, NULL, &cursor, 0), "db_cursor"))
		return false;

	key.data = (void *)s;
	key.ulen =
	key.size = len;
	key.flags = DB_DBT_MALLOC;

	/* value will be empty anyway */
	val.data = NULL;
	val.ulen =
	val.dlen =
	val.doff = 0;
	val.flags = DB_DBT_USERMEM | DB_DBT_PARTIAL;

	err = cursor->get(cursor, &key, &val, DB_SET_RANGE);

	if(!err && key.size == len + 1 && !memcmp(key.data, s, key.size)) {
		free(key.data);
		err = cursor->get(cursor, &key, &val, DB_NEXT);
	}

	switch(err) {
		case 0:
			*o = key.data;
			*olen = key.size;
			break;
		case DB_NOTFOUND:
			*o = NULL;
			*olen = 0;
			break;
		default:
			peachy = ok(err, "cursor->get");
	}

	cursor->close(cursor);

	return peachy;
}
	

bool boushi_cursor_init(boushi_t *b, boushi_cursor_t *c, const char *s, bool recursive) {
	bool peachy = true;
	size_t len;

	if(c)
		*c = boushi_cursor_0;
	if(!b)
		return false;
	if(!c)
		return bok("boushi_cursor_init", "cursor parameter is NULL");
	if(!s)
		return bok("boushi_cursor_init", "prefix parameter is NULL");

	c->b = b;
	c->recursive = recursive;

	c->prefix = malloc(strlen(s) + 2);
	if(!c->prefix)
		return ok(errno, "boushi_cursor_init");

	len = file2path(c->prefix, s);
	if(len) {
		c->prefix[len++] = '/';
		c->prefix[len] = '\0';
	}
	c->len = len;

	return peachy;
}

bool boushi_cursor_next(boushi_cursor_t *c, char **r) {
	boushi_t *b;
	bool peachy;
	char *prev;
	size_t len;
	DB *db;

	if(!c)
		return false;
	b = c->b;

	db = c->recursive ? b->files : b->directories;

	if(c->next) {
		prev = c->next;
		c->next = NULL;
		peachy = boushi_cursor_succ(b, db, prev, &c->next, &len);
		free(prev);
	} else {
		peachy = boushi_cursor_succ(b, db, c->prefix, &c->next, &len);
	}

	if(c->next && (len < c->len /* too small for the prefix */
		|| memcmp(c->next, c->prefix, c->len) /* doesn't match the prefix */
		|| (!c->recursive && strchr(c->next + c->len, '/')))) {
			/* this entry does not qualify for this listing */
			free(c->next);
			c->next = NULL;
	}
	if(peachy)
		*r = c->next;

	return peachy;
}
